*DECK USSIT3
      SUBROUTINE USSIT3(MAXNOR,NGRP,MASKG,IRES,IPLI0,IPTRK,IFTRAK,CDOOR,
     1 IMPX,NBMIX,NREG,NUN,NL,IPHASE,MAXST,MAT,VOL,KEYFLX,LEAKSW,IREX,
     2 SIGGAR,TITR,ICORR,NIRES,NBNRS,CONR,GOLD,IPPT1,IPPT2,VOLMER,UNGAR)
*
*-----------------------------------------------------------------------
*
*Purpose:
* Compute the snapshot weights as required by the resonance spectrum
* expansion (RSE) method:
* a) assume a single resonant isotope;
* b) use the standard solution doors of Dragon.
*
*Copyright:
* Copyright (C) 2023 Ecole Polytechnique de Montreal
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version
*
*Author(s): A. Hebert
*
*Parameters: input
* MAXNOR  maximum number of base points.
* NGRP    number of energy group.
* MASKG   energy group mask pointing on self-shielded groups.
* IRES    index of the resonant isotope.
* IPLI0   pointer to the internal microscopic cross section library
*         builded by the self-shielding module.
* IPTRK   pointer to the tracking (L_TRACK signature).
* IFTRAK  file unit number used to store the tracks.
* CDOOR   name of the geometry/solution operator.
* IMPX    print flag (equal to zero for no print).
* NBMIX   number of mixtures in the internal library.
* NREG    number of regions.
* NUN     number of unknowns in the flux or source vector in one
*         energy group and one band.
* NL      number of Legendre orders required in the calculation
*         (NL=1 or higher).
* IPHASE  type of flux solution (=1 use a native flux solution door;
*         =2 use collision probabilities).
* MAXST   maximum number of fixed point iterations for the ST scattering
*         source.
* MAT     index-number of the mixture type assigned to each volume.
* VOL     volumes.
* KEYFLX  pointers of fluxes in unknown vector.
* LEAKSW  leakage switch (LEAKSW=.TRUE. if neutron leakage through
*         external boundary is present).
* IREX    fuel region index assigned to each mixture. Equal to zero
*         in non-resonant mixtures or in mixtures not used.
* SIGGAR  macroscopic x-s of the non-resonant isotopes in each mixture:
*         (*,*,*,1) total; (*,*,*,2) transport correction; 
*         (*,*,*,3) P0 scattering.
* TITR    title.
* ICORR   mutual resonance shielding flag (=1 to suppress the model
*         in cases it is required in LIB operator).
* NIRES   exact number of correlated resonant isotopes.
* NBNRS   number of correlated fuel regions.
* CONR    number density of the resonant isotopes.
* GOLD    type of self-shielding model (=1.0 physical probability
*         tables; =-1001.0 resonance spectrum expansion method).
* IPPT1   pointer to LCM directory of each resonant isotope.
* IPPT2   information related to each resonant isotope:
*         IPPT2(:,1) index of a resonant region (used with infinite
*         dilution case);
*         IPPT2(:,2:4) alias name of resonant isotope.
* VOLMER  volumes of the resonant regions.
*
*Parameters: output
* UNGAR   averaged fluxes per volume.
*
*-----------------------------------------------------------------------
*
      USE GANLIB
*----
*  SUBROUTINE ARGUMENTS
*----
      TYPE(C_PTR) IPLI0,IPTRK,IPPT1(NIRES)
      INTEGER MAXNOR,NGRP,IRES,IFTRAK,IMPX,NBMIX,NREG,NUN,NL,IPHASE,
     1 MAXST,MAT(NREG),KEYFLX(NREG),IREX(NBMIX),ICORR,NIRES,NBNRS,
     2 IPPT2(NIRES,4)
      REAL VOL(NREG),SIGGAR(NBMIX,0:NIRES,NGRP,3),
     1 CONR(NBNRS,NIRES),GOLD(NIRES,NGRP),VOLMER(0:NBNRS),
     2 UNGAR(NREG,NIRES,NGRP)
      CHARACTER CDOOR*12,TITR*72
      LOGICAL LEAKSW,MASKG(NGRP)
*----
*  LOCAL VARIABLES
*----
      REAL ERR1,ERR2
      DOUBLE PRECISION T1
      CHARACTER CBDPNM*12,TEXT12*12,HSMG*131
      LOGICAL LEXAC,REBFLG
      TYPE(C_PTR) JPLI0,KPLI0,IPLIB,JPLIB1,JPLIB2,KPLIB,IPSYS,KPSYS,
     1 IOFSET,IPMACR,IPSOU
*----
*  ALLOCATABLE ARRAYS
*----
      INTEGER, ALLOCATABLE, DIMENSION(:) :: NJJ,NPSYS,MRANK
      REAL, ALLOCATABLE, DIMENSION(:) :: SIGTXS,SIGS0X
      REAL, ALLOCATABLE, DIMENSION(:,:) :: FUN,SUN
      REAL, ALLOCATABLE, DIMENSION(:,:,:,:) :: XFLUX
      DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: FLNEW
      TYPE VECTOR_ARRAY
        DOUBLE PRECISION, POINTER, DIMENSION(:) :: VECTOR
      END TYPE VECTOR_ARRAY
      TYPE MATRIX_ARRAY
        DOUBLE PRECISION, POINTER, DIMENSION(:,:) :: MATRIX
      END TYPE MATRIX_ARRAY
      TYPE(VECTOR_ARRAY), ALLOCATABLE, DIMENSION(:) :: SIGT_V,T_V
      TYPE(MATRIX_ARRAY), ALLOCATABLE, DIMENSION(:) :: PHI_M,PSI_M,
     1 SIGP
      TYPE(MATRIX_ARRAY), ALLOCATABLE, DIMENSION(:,:) :: SIGS_M
*----
*  SCRATCH STORAGE ALLOCATION
*----
      ALLOCATE(NJJ(NGRP),NPSYS(MAXNOR*NGRP),MRANK(NGRP))
      ALLOCATE(XFLUX(NBNRS,MAXNOR,NL,NIRES),PHI_M(NGRP),SIGT_V(NGRP),
     1 SIGS_M(NGRP,NGRP),T_V(NGRP),PSI_M(NGRP),SIGP(NGRP))
*----
*  FIND THE NUMBER OF COMPONENTS REQUIRED AND ALLOCATE THE LIST OF
*  ASSEMBLY MATRICES.
*----
      IPLIB=IPPT1(IRES)
      CALL LCMLEN(IPLIB,'NOR',ILONG,ITYLCM)
      IF(ILONG.NE.NGRP) THEN
        CALL LCMLIB(IPLIB)
        CALL XABORT('USSIT3: RANK ARRAY MISSING.')
      ENDIF
      CALL LCMGET(IPLIB,'NOR',MRANK)
      CALL LCMGET(IPLIB,'SVDEPS',SVDEPS)
      CALL LCMGET(IPLIB,'NJJS00',NJJ)
      JPLIB1=LCMGID(IPLIB,'GROUP-RSE')
      JPLIB2=LCMGID(IPLIB,'SIGS_M')
      NASM=0
      DO IG=1,NGRP
        IF(MASKG(IG).AND.(GOLD(IRES,IG).EQ.-1001.)) THEN
          NASM=NASM+MRANK(IG)
        ENDIF
      ENDDO
      IF(NASM.EQ.0) RETURN
*----
*  CREATE A SPECIFIC DIRECTORY FOR IRES-TH RESONANT ISOTOPE.
*----
      WRITE(CBDPNM,'(3HCOR,I4.4,1H/,I4.4)') IRES,NIRES
      CALL LCMSIX(IPLI0,CBDPNM,1)
      JPLI0=LCMGID(IPLI0,'GROUP_INFO')
      IPSYS=LCMLID(IPLI0,'ASSEMB_RSE',NASM)
      CALL LCMSIX(IPLI0,' ',2)
*----
*  LOOP OVER ENERGY GROUPS FOR THE ASSEMBLY CALCULATION.
*----
      IPOS=1
      IASM=0
      DO IG=1,NGRP
        IF(.NOT.MASKG(IG).OR.(GOLD(IRES,IG).NE.-1001.)) CYCLE
        IF(IMPX.GE.9) WRITE(6,'(22H USSIT3: energy group=,I8)') IG
        MI=MRANK(IG)
*----
*  RECOVER RSE INFORMATION FROM MICROLIB
*----
        KPLIB=LCMGIL(JPLIB1,IG)
        CALL LCMLEN(KPLIB,'SIGT_V',ILONG,ITYLCM)
        IF(ILONG.GT.MAXNOR) CALL XABORT('USSIT3: MAXNOR OVERFLOW.')
        CALL LCMGPD(KPLIB,'SIGT_V',IOFSET)
        CALL C_F_POINTER(IOFSET,SIGT_V(IG)%VECTOR,(/MI/))
        CALL LCMGPD(KPLIB,'T_V',IOFSET)
        CALL C_F_POINTER(IOFSET,T_V(IG)%VECTOR,(/MI/))
        DO JG=IG-NJJ(IG)+1,IG
          MJ=MRANK(JG)
          CALL LCMGPL(JPLIB2,IPOS+IG-JG,IOFSET)
          CALL C_F_POINTER(IOFSET,SIGS_M(IG,JG)%MATRIX,(/MI,MJ/))
        ENDDO
        IPOS=IPOS+NJJ(IG)
*----
*  COMPUTE GROUPWISE MACROSCOPIC CROSS SECTIONS.
*----
        ALLOCATE(SIGTXS(0:NBMIX),SIGS0X(0:NBMIX))
        DO IM=1,MI
          SIGTXS(0:NBMIX)=0.0
          SIGS0X(0:NBMIX)=0.0
          DO IBM=1,NBMIX
            IND=IREX(IBM)
            DENSIT=CONR(IND,IRES)
            DO 70 JRES=0,NIRES
            IF(JRES.EQ.0) THEN
*             ADMIXED NON-RESONANT ISOTOPES.
              SIGTXS(IBM)=SIGTXS(IBM)+(SIGGAR(IBM,0,IG,1)-
     1        SIGGAR(IBM,0,IG,2))
              SIGS0X(IBM)=SIGS0X(IBM)-SIGGAR(IBM,0,IG,2)
            ELSE IF((JRES.NE.IRES).AND.(IND.GT.0).AND.(ICORR.EQ.0)) THEN
*             MUTUAL SHIELDING MODEL OF CORRELATED RESONANT ISOTOPES.
              CALL XABORT('USSIT3: CORRELATION MODEL NOT AVAILABLE(1).')
            ENDIF
   70       CONTINUE
            IF(IND.GT.0) THEN
              SIGTXS(IBM)=SIGTXS(IBM)+DENSIT*REAL(SIGT_V(IG)%VECTOR(IM))
              SIGS0X(IBM)=SIGS0X(IBM)+DENSIT*
     1               REAL(SIGS_M(IG,IG)%MATRIX(IM,IM))
            ENDIF
          ENDDO
          NPSYS(IASM+IM)=IASM+IM
          KPSYS=LCMDIL(IPSYS,IASM+IM)
          CALL LCMPUT(KPSYS,'DRAGON-TXSC',NBMIX+1,2,SIGTXS(0))
          CALL LCMPUT(KPSYS,'DRAGON-S0XSC',NBMIX+1,2,SIGS0X(0))
        ENDDO
        IASM=IASM+MI
        DEALLOCATE(SIGS0X,SIGTXS)
      ENDDO
*----
*  ASSEMBLY MATRIX OR REDUCED COLLISION PROBABILITIES CALCULATION.
*----
      ISTRM=1
      NANI=1
      NW=0
      KNORM=1
      NALBP=0
      IMPY=MAX(0,IMPX-3)
      IF(IPHASE.EQ.1) THEN
*        USE A NATIVE DOOR.
         CALL DOORAV(CDOOR,IPSYS,NPSYS,IPTRK,IFTRAK,IMPY,NASM,NREG,
     1   NBMIX,NANI,NW,MAT,VOL,KNORM,LEAKSW,TITR,NALBP,ISTRM)
      ELSE IF(IPHASE.EQ.2) THEN
*        USE A COLLISION PROBABILITY DOOR.
         IPIJK=1
         CALL DOORPV(CDOOR,IPSYS,NPSYS,IPTRK,IFTRAK,IMPY,NASM,NREG,
     1   NBMIX,NANI,MAT,VOL,KNORM,IPIJK,LEAKSW,.FALSE.,TITR,NALBP)
      ENDIF
*----
*  LOOP OVER ENERGY GROUPS FOR THE FLUX CALCULATION.
*----
      IASM=0
      DO IG=1,NGRP
        MI=MRANK(IG)
        IF(.NOT.MASKG(IG).OR.(GOLD(IRES,IG).NE.-1001.)) CYCLE
        ALLOCATE(PSI_M(IG)%MATRIX(NUN,MI))
        PSI_M(IG)%MATRIX(:NUN,:MI)=0.0D0
        ITER=0
   20   ITER=ITER+1
        IF(ITER.GT.MAXST) CALL XABORT('USSIT3: TOO MANY ITERATIONS.')
        ERR1=0.0
        ERR2=0.0
*----
*  COMPUTE THE AVERAGED SOURCE.
*----
        ALLOCATE(FUN(NUN,MI),SUN(NUN,MI))
        SUN(:NUN,:MI)=0.0
        DO IM=1,MI
          KPSYS=LCMGIL(IPSYS,IASM+IM)
          CALL LCMLEN(KPSYS,'FUNKNO$USS',ILENG,ITYLCM)
          IF(ILENG.EQ.NUN) THEN
            CALL LCMGET(KPSYS,'FUNKNO$USS',FUN(1,IM))
          ELSE
            FUN(:NUN,IM)=0.0
          ENDIF
          SUN(:NUN,IM)=0.0D0
          NPSYS(IM)=IASM+IM
          DO I=1,NREG
            IBM=MAT(I)
            IF(IBM.EQ.0) CYCLE
            IUN=KEYFLX(I)
            IND=IREX(IBM)
            DENSIT=CONR(IND,IRES)
            T1=0.0D0
            DO JRES=0,NIRES
              IF(JRES.EQ.0) THEN
                T1=T1+SIGGAR(IBM,0,IG,3)*T_V(IG)%VECTOR(IM)
              ELSE IF((JRES.EQ.IRES).AND.(IND.GT.0)) THEN
                DO JG=IG-NJJ(IG)+1,IG
                  MJ=MRANK(JG)
                  DO JM=1,MJ
                    IF((JM.EQ.IM).AND.(JG.EQ.IG)) CYCLE
                    T1=T1+DENSIT*SIGS_M(IG,JG)%MATRIX(IM,JM)*
     1              PSI_M(JG)%MATRIX(IUN,JM)
                  ENDDO
                ENDDO
              ELSE IF((JRES.NE.IRES).AND.(IND.GT.0).AND.(ICORR.EQ.0))
     1        THEN
*               MUTUAL SHIELDING MODEL OF CORRELATED RESONANT ISOTOPES.
                CALL XABORT('USSIT3: CORRELATION MODEL NOT AVAILABLE.')
              ELSE IF((JRES.NE.IRES).AND.(IND.GT.0).AND.(ICORR.EQ.1))
     1        THEN
                T1=T1+SIGGAR(IBM,0,IG,3)*T_V(IG)%VECTOR(IM)
              ENDIF
            ENDDO
            SUN(IUN,IM)=REAL(T1,4)
          ENDDO
        ENDDO
*----
*  SOLVE FOR THE MULTIBAND FLUX.
*----
        IDIR=0
        LEXAC=.FALSE.
        IPMACR=C_NULL_PTR
        IPSOU=C_NULL_PTR
        REBFLG=.FALSE.
        CALL DOORFV(CDOOR,IPSYS,NPSYS,IPTRK,IFTRAK,IMPX,MI,NBMIX,IDIR,
     1  NREG,NUN,IPHASE,LEXAC,MAT,VOL,KEYFLX,TITR,SUN,FUN,IPMACR,IPSOU,
     2  REBFLG)
*----
*  CONVERGENCE CONTROL.
*----
        DO IM=1,MI
          KPSYS=LCMGIL(IPSYS,IASM+IM)
          CALL LCMPUT(KPSYS,'FUNKNO$USS',NUN,2,FUN(1,IM))
          DO I=1,NREG
            IUN=KEYFLX(I)
            DELTA=REAL(FUN(IUN,IM)-PSI_M(IG)%MATRIX(IUN,IM),4)
            ERR1=MAX(ERR1,ABS(DELTA))
            ERR2=MAX(ERR2,ABS(FUN(IUN,IM)))
          ENDDO
          PSI_M(IG)%MATRIX(:NUN,IM)=FUN(:NUN,IM)
        ENDDO
        DEALLOCATE(SUN,FUN)
        IF(IMPX.GT.2) THEN
          WRITE(TEXT12,'(3A4)') (IPPT2(IRES,I),I=2,4)
          WRITE(6,'(15H USSIT3: GROUP=,I5,15H. RSE ITERATION,I4,
     1    11H. ISOTOPE='',A12,9H''. ERROR=,1P,E11.4,1H.)') IG,
     2    ITER,TEXT12,ERR1
        ENDIF
        IF(ERR1.GT.1.0E-4*ERR2) GO TO 20
        IF(IMPX.GT.1) THEN
           WRITE(TEXT12,'(3A4)') (IPPT2(IRES,I),I=2,4)
           WRITE(6,'(15H USSIT3: GROUP=,I5,24H. RSE ITERATION CONVERGE,
     1     6HNCE IN,I4,22H ITERATIONS. ISOTOPE='',A12,2H''.)') IG,
     2     ITER,TEXT12
        ENDIF
*----
* RECOVER RSE INFORMATION FROM MICROLIB
*----
        KPLIB=LCMGIL(JPLIB1,IG)
        CALL LCMLEN(KPLIB,'RSE-TABLE',ILONG,ITYLCM)
        NPART=ILONG/MI
        CALL LCMGPD(KPLIB,'RSE-TABLE',IOFSET)
        CALL C_F_POINTER(IOFSET,SIGP(IG)%MATRIX,(/NPART,MI/))
*----
* COMPUTE SNAPSHOT WEIGHTS AND HOMOGENIZE THE FLUX
*----
        ALLOCATE(FLNEW(NBNRS,MI))
        UNGAR(:NREG,IRES,IG)=0.0
        FLNEW(:NBNRS,:MI)=0.0D0
        DO I=1,NREG
          IF(MAT(I).EQ.0) CYCLE
          IUN=KEYFLX(I)
          IND=IREX(MAT(I))
          DO IM=1,MI
           IF(IND.GT.0) FLNEW(IND,IM)=FLNEW(IND,IM)+
     1           PSI_M(IG)%MATRIX(IUN,IM)*VOL(I)
           UNGAR(I,IRES,IG)=UNGAR(I,IRES,IG)+REAL(SIGP(IG)%MATRIX(1,IM)*
     1     PSI_M(IG)%MATRIX(IUN,IM),4)
          ENDDO
        ENDDO
        DO IM=1,MI
          DO IND=1,NBNRS
            XFLUX(IND,IM,1,IRES)=REAL(FLNEW(IND,IM)/VOLMER(IND))
          ENDDO
        ENDDO
        DEALLOCATE(FLNEW)
        IF(IMPX.GT.2) THEN
          DO IND=1,NBNRS
            T1=0.0D0
            DO IM=1,MI
              T1=T1+SIGP(IG)%MATRIX(1,IM)*XFLUX(IND,IM,1,IRES)
            ENDDO
            WRITE(6,'(31H USSIT3: AVERAGED FLUX IN GROUP,I4,9H AND RESO,
     1      11HNANT REGION,I4,21H FOR RESONANT ISOTOPE,I4,2H =,F9.5)')
     2      IG,IND,IRES,T1
          ENDDO
        ENDIF
        KPLI0=LCMGIL(JPLI0,IG)
        CALL LCMPUT(KPLI0,'NWT0-PT',NBNRS*MAXNOR*NL,2,XFLUX(1,1,1,IRES))
        IASM=IASM+MI
      ENDDO
*----
*  USE IL=1 VALUES FOR XFLUX AT HIGHER LEGENDRE ORDERS.
*----
      DO IL=2,NL
        DO IG=1,NGRP
          IF(.NOT.MASKG(IG).OR.(GOLD(IRES,IG).NE.-1001.)) CYCLE
          MI=MRANK(IG)
          IF(MI.EQ.0) CYCLE
          KPLI0=LCMGIL(JPLI0,IG)
          CALL LCMLEN(KPLI0,'NWT0-PT',ILONG,ITYLCM)
          IF(ILONG.GT.NBNRS*MAXNOR*NL) THEN
            WRITE(TEXT12,'(3A4)') (IPPT2(IRES,J0),J0=2,4)
            WRITE(HSMG,'(34HUSSIT3: FLUX OVERFLOW FOR ISOTOPE ,A12)')
     1      TEXT12
            CALL XABORT(HSMG)
          ENDIF
          CALL LCMGET(KPLI0,'NWT0-PT',XFLUX(1,1,1,IRES))
          DO IM=1,MI
            DO IND=1,NBNRS
              XFLUX(IND,IM,IL,IRES)=XFLUX(IND,IM,1,IRES)
            ENDDO
          ENDDO
          CALL LCMPUT(KPLI0,'NWT0-PT',NBNRS*MAXNOR*NL,2,
     1    XFLUX(1,1,1,IRES))
        ENDDO
      ENDDO
*----
*  SCRATCH STORAGE DEALLOCATION
*----
      DO IG=1,NGRP
        IF(GOLD(IRES,IG).NE.-1001.) CYCLE
        DEALLOCATE(PSI_M(IG)%MATRIX)
      ENDDO
      DEALLOCATE(SIGP,PSI_M,T_V,SIGS_M,SIGT_V,PHI_M,XFLUX)
      DEALLOCATE(MRANK,NPSYS,NJJ)
      RETURN
      END
